// GIHarness.java
//---------------

package cw.GI;


import cw.log.Log;

/**
 *
 * This class is the generic part of the Game Harness.<br>
 * Once implemented for a specific platform, it will
 * have the following functionalities:<br>
 * 
 *
 * <ul><li>
 * Keyboard support<br> 
 * maintains an array of keys.<br>
 * provide functionality for current key status reading with isKeyPressed()
 * and recent changes with wasKeyPressed()
 *
 * </li><li>
 * Time monitoring<br>
 * runs a thread with smooth calling (i.e. the game runs at constant speed)<br>
 *
 * </li><li>
 * Graphics support<br>
 * Loading images<br>
 * Printing images<br>
 * Printing rectangle<br>
 * Printing text<br>
 * </li></lu>
 *
 */



abstract public class GIHarness
{
  public int mMaxX=0; // holds the internal screen size
  public int mMaxY=0;


  //-- Misc stuff ----------------------------------------------



  /**
   * The number of images available. must be changed only before init() 
   */
  public int mImagesAvailable=200;
  
  public boolean mErrorFlag;
  public String mErrorMsg;
  
 
  public void init()
  {
   Log.msg('I',"GIHarness.init() ");
    
    if (mMaxX==0) throw new RuntimeException("setSize() was not called");
    
    initKeys();
    initImages(mImagesAvailable);
    mErrorFlag=false;
  
    
    mExitFlag=false;
    mThread=new GIThread(this);
   	// the thread is started in postStart
    
  }
  
  /**
   * causes the game to freeze and display an error message
   *
   */
  public void setError(String msg)
  {
   Log.msg('E',"setError() msg=",msg);
    mErrorFlag=true;
    mErrorMsg=msg;
  }

  /**
   *
   * set the screen size. should be called before init()
   */

  public void setSize(int x,int y)
  {
    
    
    // Trick to switch off logging when running on C60
    // TODO: remove next line
    //Log.mLogOnScreen=y>100;
    
   Log.msg('I',"GIHarness.setSize() x=",x,"y=",y);
    
    mMaxX=x;
    mMaxY=y;
    
  }
  
  
  
  /**
   * destructor . frees resources
   */
  public void done()
  {
   Log.msg('I',"GIHarness.done()");
    for (int i=0;i<mImages.length;i++)
    {
      mImages[i]=null;
    }
    if (mGameMgr!=null)
    {
      mGameMgr.done();
      mGameMgr=null;
    }
    mThread=null;
  }
  
  public GIGameMgr mGameMgr;
  
  public void setGameMgr(GIGameMgr pGameMgr)
  {
    Log.msg('I',"GIHarness.setGameMgr()");
     mGameMgr=pGameMgr;
     mGameMgr.mH=this;
  }
  
  //-- Key stuff ----------------------------------------------

  /**
   * mKeyCodeXXX are the actual key code for these game actions.
   * these codes are platform specific and must be initialised by the
   * platform implementation.  
   *
   * default values are for PC keyboard.
   */
   
   
  public int mKeyCodeUp=38;
  public int mKeyCodeDown=40;
  public int mKeyCodeLeft=37;
  public int mKeyCodeRight=39;
  
  public int mKeyCodeUpRight='I';
  public int mKeyCodeUpLeft='O';
  public int mKeyCodeDownRight='K';
  public int mKeyCodeDownLeft='L';
  
  
  public int mKeyCodeFire=32;
  public int mKeyCodeA='Z';
  public int mKeyCodeB='X';
  public int mKeyCodeC='C';
  public int mKeyCodeD='V';
  
  
  
           

  /**
   * array of keys status
   */
  public int mKeyCodes[]; 
  //public int mLastReleasedKeyCode;
  
  
  /**
   * internal clock for keys.
   * has to be bigger than mKeyPressedMemDelay
   */
  public int mKeyCnt; // key clock cnt
  
  /*
   * parameter delay for forgetting a key has been pressed 
   * but is status was never checked wit wasKeyPressed()
   */
  public final static int mKeyPressedMemDelay=10;
  
  /**
   * initialise the key codes
   */
  public void initKeys()
  //----------------
  {
    
   Log.msg('I',"GIHarness.initKeys()");
    mKeyCodes =new int[256];
    //mLastReleasedKeyCode=-1;
    
    mKeyCnt=mKeyPressedMemDelay+1; // has to be bigger than mKeyPressedMemDelay
     
    for (int i=0;i<256;i++)
    {
        mKeyCodes[i]=0;
    }
   
  }
  
 
   
  /**
   * the platform implementation should call this function
   * when a key is pressed
   */
  public void oneKeyPressed(int key) 
  {
    key=key>=0?(key):(256+key);
   //Nog.msg('D',"GIHarness.key pressed:",key);
    mKeyCodes[key]=-1;
    
  } 
 
  /**
   * the  platform implementation should call this function
   * when a key is released
   */
    
  public void oneKeyReleased(int key) 
  {
    key=key>=0?(key):(256+key);
   //Nog.msg('D',"GIHarness.key released:",key);
    mKeyCodes[key]=mKeyCnt;
   // mLastReleasedKeyCode=key;
  }


  /**
   * update the key clock cnt
   *
   */
  public void updateKeyCodes()
  {
      mKeyCnt++;
  //Nog.msg('D',"updateKeyCodes() mKeyCnt=",mKeyCnt);    
  }
  
  
  /**
   * check the current status of a key.
   *
   */
  public boolean isKeyPressed(int key)
  {
      key=key>=0?(key):(256+key);
  //Nog.msg('D',"isKeyPressed(",key,")");
      boolean retVal=mKeyCodes[key]==-1;
  //Nog.msg('D',"isKeyPressed(",key,") returns ",retVal);
    return retVal;
  }
  
  
  
  /**
   *
   * check if a key was pressed recently.
   * if this is the case, returns true, and clear the key status. Thus next calls
   * of wasKeyPressed for this key will return false, until the key is pressed again.
   *
   */

    
  public boolean wasKeyPressed(int key)
  {
      key=key>=0?(key):(256+key);
      boolean retVal=mKeyCodes[key]>mKeyCnt-mKeyPressedMemDelay;
  //Nog.msg('D',"wasKeyPressed(",key,") returns ",retVal," mKeyCodes[key]=",mKeyCodes[key]," mKeyCnt-1=",mKeyCnt-1);
      if (retVal) mKeyCodes[key]=0;
    return retVal;
  }
  
  
  //-- Image box stuff ------------------------------------
    
  
 
       
  public GIImageBox[] mImages;
  
  /**
   * initialise the image array.
   */
  public void initImages(int pSize)
  {
    mImages=new GIImageBox[pSize];
    for (int i=0;i<mImages.length;i++)
      mImages[i]=null;
  }
  
  /**
   * load an image at specified index into a read only buffer
   */
  public boolean loadImage(int index,String name,String pExt)
  {
    return loadImage(index,name,pExt,false);    
  }
  
  /**
   * load from disk an image at specified index.
   *
   * @param forWrite used to specify whether it will be possible to draw into the image buffer will be write or the image buffer will be read-only
   */
  public boolean loadImage(int index,String name,String pExt,boolean forWrite)
  {
    boolean retVal=false;
    
    GIImageBox imageBox;

   //Nog.msg('T',"GIHarness.loadImage(",name+pExt,") start for image at index",index);    
    try
    {
      imageBox=loadImageBox(name,pExt,forWrite);
    }
    catch (Exception e)
    {
      imageBox=null;
    }
    
    if (imageBox!=null)
    {
    //DEBUG{
      if (mImages[index] !=null) throw new RuntimeException("loading image "+name+pExt+". at index "+index+". there is already another image at this index");
    //DEBUG}  
      mImages[index]=imageBox;
      retVal=true;
      
     Log.msg('I',"image ",name+pExt," size: ",mImages[index].mX,"x",mImages[index].mY);
    }
    else
    {
       Log.msg('E',"GIHarness.loadImage(",name+pExt,") failed");
  
    }
    
    return retVal;
  }
  
  
   
  /**
   * load from disk a bunch of pSixe images starting at index pStart.
   *
   */
  public void loadImages(int pStart,int pSize,String pName,String pExt) 
  //---------------------------------------------------------------------
  {
    for (int i=0;i<pSize;i++)
    {
       String n=""+i;
       if (i<10) n="0"+n;
       
       loadImage(pStart+i,pName+n,pExt,false);
    }
  }
  
  /**
   *
   * create a new image with writing abilities
   */
  public boolean createImage(int index,int x,int y)
  {
    GIImageBox lImageBox=createImageBox(x,y);
    if (lImageBox!=null)
    {
        mImages[index]=lImageBox;
        return true;
    }    
    else return false; 
  }
  
  
  /**
   * print an image into the default output image
   */
   
  public void printImage(int x,int y,int image)
  {
      printImage(x,y,image,mDefaultWriteImage);
  }
  
  
  
  /**
   * platform specific function required by createImage() 
   */
  abstract public GIImageBox createImageBox(int x,int y);
  
  
  
  /**
   * platform specific function required by loadImage() 
   */
  abstract public GIImageBox loadImageBox(String name,String pExt,boolean AllowWriting) throws Exception;
 
 
 
  /**
   * platform specific function required by printImage() 
   */
  abstract public void printImage(int x,int y,int image,int toImage);
  
  
  // text
    
 
  /**
   * text styles.
   */ 
  public final int ROUNDED_FONT=-2;
  public final int BOXED_FONT=-1;
  public final int BOLD_FONT=2;
  public final int PLAIN_FONT=0;
  
  
  /**
   * write a text on one line.
   */
  public void writeTextLine(int x,int y,String text,int color,int style,int toImage)
  {
  	  int ucolor=0xffffff-color;
  	  
  	 
  	  
  	  
  	  switch (style)
  	  { 
  	    case BOXED_FONT:
      	  printRect(x,y,x+text.length()*mFontSizeX,y+mFontSizeY,ucolor,true,toImage);
          writeRawTextLine(x,y+mFontSizeY,text,color,toImage);
          break;
        case ROUNDED_FONT:
          writeRawTextLine(x-1,y+mFontSizeY-1,text,ucolor,toImage);
          writeRawTextLine(x-1,y+mFontSizeY+1,text,ucolor,toImage);
          writeRawTextLine(x+1,y+mFontSizeY-1,text,ucolor,toImage);
          writeRawTextLine(x+1,y+mFontSizeY+1,text,ucolor,toImage);
          writeRawTextLine(x,y+mFontSizeY,text,color,toImage);
          break;
        case BOLD_FONT:
        case PLAIN_FONT:
          for (int i=0;i<=style;i++)
          {
            writeRawTextLine(x+i,y+mFontSizeY+i,text,color,toImage);
          }
          break;
      }
  }    
  
  int mDefaultWriteImage=mImagesAvailable-1;
  
  public void writeText(int x,int y,String pText,int color,int style)
  {
      writeText(x,y,pText,color,style,mDefaultWriteImage);
  }
  
  public int mFontSizeY=10;
  public int mFontSizeX=8;
  
  public void writeText(int x,int y,String pText,int color,int style,int toImage)
  //-------------------
  {
 
    
      int lThickness=style;
      
      int curPos=0;
      int lLine=0;
 
      
      do
      {
        int newLine=pText.indexOf('\n',curPos);
      
        if (newLine==-1) newLine=pText.length();
        String subString=pText.substring(curPos,newLine);
        
        writeTextLine(x,y+lLine*mFontSizeY,subString,color,style,toImage);
        
        curPos=newLine+1;
        lLine++;
      
      } while (curPos<pText.length());
    
  }  
  
  
  abstract public void writeRawTextLine(int x,int y,String text,int color,int toImage);  
 
  //--+ rectangle +-------------------------------------------------------------
  
 
  public void printRect(int x1,int y1,int x2,int y2,int color,boolean filled)
  {
      printRect(x1, y1, x2, y2, color,filled,mDefaultWriteImage);
  }
 
  abstract public void printRect(int x1,int y1,int x2,int y2,int color,boolean filled,int toImage);
 
 
 //--+ paint flow +---------------------------------------------------------------------------
  
  
  
  // double buffering check
  //boolean mReadyForPaint=false;
   
 // public void bufferReset()
  //{
    //mReadyForPaint=false;
    //processBackgroundScroll() ;
    //bufferClear();
  //}
  
  
  
  
  
  //public void bufferReadyForPaint()
 // {
   
   // mReadyForPaint=true;
   // if (mPaintRequestWhenReady) 
   // {
   //    requestPaint();
   //  }
  
 // }
  
 // boolean mPaintRequestWhenReady=false;
  

  
  
  //{
  //  throw new RuntimeException("method not implemented");
  // }
          
 // boolean mInPaintWrap=false;
 // boolean mCanPaint=false;
  
  
  /**
   * used to indicate to the harness that
   *
   */
 // public void canPaint(boolean pCanPaint)
  //{
  //  mCanPaint=pCanPaint;
    
  //}
  /**
   * this function is called by the external side harness
   * to perform the paint
   *
   * paint will be performed only if the thread is not in step()
   */
  public void paintWrap()
  {
   //Nog.msg('T',"paintWrap() ");
    try
    
    
    // if (mCanPaint)
    {
    
    
      if (mErrorFlag)
      {
        
        
        printRect(0,0,mMaxX,mMaxY,(int)(0xFFFFFF-mReportColor),true);
        writeText(0,0,mErrorMsg+"\n<"+mDisplayReportKey+">",(int)mReportColor,PLAIN_FONT); 
        if (wasKeyPressed(mDisplayReportKey)) mErrorFlag=false;
        
       }
      else synchronized (mThread.mSemaphore)
      {
      
        //mInPaintWrap=true;
        //long timeBefore=System.currentTimeMillis();
        
        //    bufferReset();
        paint();
        //bufferReadyForPaint();
        
        
        
        // flash management
        if (mFlash)
        {
          if (mThread.mClockCnt%2==0) printRect(0,0,mMaxX,mMaxY,mFlashColor,true);
          if (--mFlashCnt==0) mFlash=false;  
        }
        
        
        if (wasKeyPressed(mDisplayReportKey)) showReport(mShowReport+1);
        if (mShowReport>0) displayReport();
        
        //long timeAfter=System.currentTimeMillis();
        // mThread.addStatWorkTime(timeAfter-timeBefore);
        
       // mInPaintWrap=false;
      }
     // else 
     // {
     //   mThread.mStatCanceledAddWord++;
       //Nog.msg('T',"paint() cancelled. cancelled cnt=",    mThread.mStatCanceledAddWord);
     // }
          
    }   
    catch (Exception e)
    {
      Log.msge('E',"got exception ",e);
      setError("in paintWrap. ex:\n"+e.getClass().getName()+"\n"+e.getMessage());
      //mInPaintWrap=false;
    } 
  }

  
  
  
  abstract public void requestPaint();
  //abstract public void bufferClear();
 
  
  
  public boolean mDisplayProgress=true;
  public int mProgress=0;
  
 /**
  * requests the game mgr to do its paintings
  */
   public void paint()
   {
     if(!mDisplayProgress) mGameMgr.paint();
     else
     {
     	printRect(0,0,mMaxX,mMaxY,(int)(0xFFFFFF),true);
        writeText(10,50,"Starting...",(int)0x000000,PLAIN_FONT); 
        printRect(10,40,mMaxX-9,45,(int)(0x000000),false);
        printRect(11,41,10+(mProgress*(mMaxX-20)/100),45,(int)(0x0000FF),true);
     }
   }
  
  
 //-- smooth running -------------------------------------- 
  
  boolean mExitFlag; // the thread will die when this flag is true
  public GIThread mThread; // reference to the thread
 
 
  /**
   *
   * launch the game thread. after this function the 
   * step() function will be called at regular intervals
   */
  
  public int mStartCnt=0;
  public void startInc()
  {
   Log.msg('I',"GIHarness.startInc() mStartCnt=",mStartCnt,". started");
   
   mGameMgr.init(mStartCnt);
   mStartCnt++;
   mProgress++;
   		
   Log.msg('I',"GIHarness.start() done. with mStartCnt ",mStartCnt-1,". mStartCnt incremented");
  }
  
  public boolean started()
  {
   boolean retVal=mGameMgr.inited();
   if (retVal && !mStarted) 
   	{
   	 	postStarted();
   	 	new Thread(mThread).start();
  	    mStarted=true;
  	    mDisplayProgress=false;
   	}
   return retVal;
  }
  

  private boolean mStarted=false;
  public void postStarted()
  {
  	 // nothing to do now
  	 // after this function the GIThread will start 
  }
  
   
  public void start()
  {
  	while(!started())
  	{
  		startInc();
  		requestPaint();
  		/*try
		{
  			Thread.sleep(1000);
		} catch (Exception e)
		{ // noting to do
		}*/
  	}	
  }


  /**
   * causes the game thread to stop,
   * which in turn calls close as last operation
   */
  public void stop()
  {
   Log.msg('I',"GIHarness.stop()");
    mExitFlag=true;
  }
  
  
  /**
   * This function will be called at regular interval
   * by the game thread
   *
   */
   public void step(int mClock)
   {
     mGameMgr.step(mClock);
   }
   
  
 //--+ report stuff +-------------------------------------------------------
  
  /**
   * pressing this key will switch the report screen 
   */
  public char mDisplayReportKey='0';
  
  /**
   * key codes for the report screen actions
   */
  public char mDisplayReportKey1='1';
  public char mDisplayReportKey2='2';
  public char mDisplayReportKey3='3';
  public char mDisplayReportKey4='4';
  
 /**
   * use this function to change report screen.
   * 
   */
  public void showReport(int pShowReport)
  {
    mShowReport=pShowReport;
   //Nog.msg('D',"showReport() mShowReport=",mShowReport);  
  }
  
  
   /**
   * holds the value of the current report screen.
   * 0 means no report screen
   */
  public int mShowReport=0;
  long mReportColor=0x0000FF; // report color
  int mReportDisplay=3; // way of displaying
  
  
  
  /**
   * displays a report on the screen.
   *
   * 0: do not display any report
   * 1: thread stats report
   *    '1' - increase delay by 10 ms
   *    '2' - decrease delay by 10 ms
   *    '3' - reset stats
   *
   * 2: screen report
   *    '1' - toggle Log.mMasterFlag
   *    '2' - change report screen display colors
   */
   
   
   
   
  public void displayReport()
  {
      if (mShowReport==0) return; // exit quickly
      String msg="";
      
      char bignoul="/-\\|/-\\|".charAt(mThread.mClockCnt%8);
   
      
    switch(mShowReport)
    
    {
      case 1: // thread report
        msg="th:"
          +mThread.mClockCnt+"\n"
          +mThread.getStatString()+"\nd="+mThread.mClockDelay
          +"\n<"+mDisplayReportKey1+mDisplayReportKey2+mDisplayReportKey3
          +">";
         if (wasKeyPressed(mDisplayReportKey1)) 
         {
            mThread.mClockDelay+=10;
         }
         if (wasKeyPressed(mDisplayReportKey2)) 
         {
            mThread.mClockDelay-=10;
            if (mThread.mClockDelay<0) mThread.mClockDelay=0;;
         }
         
         if (wasKeyPressed(mDisplayReportKey3)) mThread.initStat();
       
      
         
      break;
      
      case 2: // screen report
    
        msg="screen:"+bignoul+"\n"+mMaxX+'x'+mMaxY
         +"\n<"+mDisplayReportKey1+(Log.mMasterFlag?'+':'-')
          +mDisplayReportKey2+mReportDisplay
          +">";
        if (wasKeyPressed(mDisplayReportKey1)) Log.setMasterFlag(!Log.mMasterFlag);
        if (wasKeyPressed(mDisplayReportKey2)) 
        {
            switch (++mReportDisplay) 
            {
              case 1:mReportColor=0xFF0000;break;
              case 2:mReportColor=0x00FF00;break;
              case 3:mReportColor=0x0000FF;break;
              
              case 4:mReportColor=0x00FFFF;break;
              
              case 5:mReportColor=0xFFFF00;break;
             
              case 6:mReportColor=0xFFFFFF;break;
              default:
                mReportDisplay=0;
                mReportColor=0;
                break;
            }
        }
        break;
        
        
   
      case 3: // key report
    
        msg="keys:"+bignoul+"\n";
        
        for (int i=0;i<256;i++)
           {
              if (isKeyPressed(i))
              {
                  msg=msg+i;
                  
                 
                  
                  msg+="\n";
 
                  
              } 
           } 
           
          if (isKeyPressed(mKeyCodeUp)) msg+=" up";            
          if (isKeyPressed(mKeyCodeDown)) msg+=" down";
          if (isKeyPressed(mKeyCodeLeft)) msg+=" left";
          if (isKeyPressed(mKeyCodeRight)) msg+=" right";
          if (isKeyPressed(mKeyCodeUpLeft)) msg+="up-left";
          if (isKeyPressed(mKeyCodeUpRight)) msg+="up-right";
          if (isKeyPressed(mKeyCodeDownLeft)) msg+="down-left";
          if (isKeyPressed(mKeyCodeDownRight)) msg+="down-right";
        
          if (isKeyPressed(mKeyCodeFire)) msg+=" fire";
          if (isKeyPressed(mKeyCodeA)) msg+=" A";
          if (isKeyPressed(mKeyCodeB)) msg+=" B";
          if (isKeyPressed(mKeyCodeC)) msg+=" C";
          if (isKeyPressed(mKeyCodeD)) msg+=" D";   
           
           
        break;
      // anything else cause report screen to be set to none.
      default: mShowReport=0;    
    }
    
    
    
    
   
    
    // now display the report screen
    if (mReportDisplay==0) printRect(0,0,mMaxX,mMaxY,(int)(0xFFFFFF-mReportColor),true);
   
    writeText(0,0,msg,(int)mReportColor,BOXED_FONT); 
    //Log.msg('I',"displayReport() msg=",msg);
    
  }


 //--+ flashing +--------------------------------------------------------------
 public boolean mFlash=false;
 public int mFlashColor=0xFFFFFF;
 public int mFlashCnt;
 
 public void startFlash(int cnt)
 {
   mFlash=true;
   mFlashCnt=cnt;
 }
  
    
    
    
 public final static int BLUE=0x0000FF;
 public final static int YELLOW=0xFFFF00;
 public final static int RED=0xFF0000;
 public final static int GREEN=0x00FF00;
 public final static int PINK=0xFF00FF;
 public final static int BLACK=0x000000;
 public final static int WHITE=0xFFFFFF;
 public final static int GRAY=0x777777;
     
    
    
} 










